from base64 import b64decode
from io import BytesIO

import numpy as n
import pytest

from cryosparc.dataset import CSDAT_FORMAT, Column
from cryosparc.dataset.row import Row

from .conftest import Dataset


@pytest.fixture
def io_data():
    data = b64decode(
        (
            b"k05VTVBZAQCmAHsnZGVzY3InOiBbKCd1aWQnLCAnPHU4JyksICgnZmllbGQxJywgJzx1NCcpLC"
            b"AoJ2ZpZWxkMicsICc8ZjQnKSwgKCdmaWVsZDMnLCAnfFM2JyksICgnZmllbGQ0JywgJzxmOCcp"
            b"LCAoJ2ZpZWxkNScsICc8aTgnKV0sICdmb3J0cmFuX29yZGVyJzogRmFsc2UsICdzaGFwZSc6IC"
            b"gyLCksIH0gIAp7AAAAAAAAACoAAADD9UhASGVsbG8AAAAAAAAAAAArAAAAAAAAAMgBAAAAAAAA"
            b"KgAAAFK4LkBXb3JsZAAAAAAAAADwPysAAAAAAAAA"
        )
    )
    return BytesIO(data)


@pytest.fixture
def small_dset():
    field3 = "long/fieldwithsuperduperlongcolumnnamethatislongandtestable"
    dset = Dataset.allocate(
        4,
        fields=[
            ("field/1", "u8", (2,)),
            ("field/2", "f4"),
            (field3, "O"),
        ],
    )
    dset["field/1"] = (42, 43)
    dset["field/2"] = n.array([3.14, 2.73, 1.62, 3.14], dtype="f8")
    dset[field3][:] = n.array(["Hello", "World", "!", "!"])
    return dset


@pytest.fixture
def small_dset_path(tmp_path, small_dset):
    path = tmp_path / "small_dset.cs"
    small_dset.save(path, format=CSDAT_FORMAT)
    return path


@pytest.fixture
def small_dset_stream(small_dset):
    stream = BytesIO()
    small_dset.save(stream, format=CSDAT_FORMAT)
    stream.seek(0)
    return stream


@pytest.fixture
def empty_dset():
    field3 = "long/fieldwithsuperduperlongcolumnnamethatislongandtestable"
    return Dataset.allocate(
        0,
        fields=[
            ("field/1", "u8", (2,)),
            ("field/2", "f4"),
            (field3, "O"),
        ],
    )


@pytest.fixture
def empty_dset_stream(empty_dset):
    stream = BytesIO()
    empty_dset.save(stream, format=CSDAT_FORMAT)
    stream.seek(0)
    return stream


def test_allocate():
    storage = Dataset.allocate(size=2000000, fields=[("field1", "u8"), ("field2", "f4"), ("field3", "O")])
    assert storage is not None


def test_populate_new_0():
    storage = Dataset.allocate(size=0)
    assert len(storage) == 0


def test_populate_new_1():
    storage = Dataset.allocate(size=1)
    assert len(storage) == 1


def test_populate_new_many():
    storage = Dataset.allocate(size=3)
    assert len(storage) == 3


def test_storage_from_other():
    storage1 = Dataset(3)
    storage2 = storage1.copy()
    assert len(storage2) == 3


def test_basic_data_constructor():
    data = Dataset()
    assert len(data) == 0
    assert len(data.descr()) == 1


def test_empty_data_constructor():
    data = Dataset(0)
    assert len(data) == 0
    assert len(data.descr()) == 1


def test_fields():
    data = Dataset.allocate(3, [("test1", "<u4"), ("test2", "<f8", (2,))])
    assert data.fields() == ["uid", "test1", "test2"]
    assert data.fields(exclude_uid=True) == ["test1", "test2"]


def test_descr():
    data = Dataset.allocate(3, [("test1", "<u4"), ("test2", "<f8", (2,))])
    expected_descr = [("uid", "<u8"), ("test1", "<u4"), ("test2", "<f8", (2,))]
    assert data.descr() == expected_descr
    assert data.descr(exclude_uid=True) == expected_descr[1:]


def test_invalid_data_fields():
    # This is ok actually
    assert Dataset(
        [
            ("uid", n.array([1, 2, 3])),
            ("dat", ["Hello", "World", "!"]),
        ]
    )


def test_uneven_data_fields():
    with pytest.raises(AssertionError):
        Dataset(
            [
                ("uid", n.array([1, 2, 3])),
                ("dat", n.array(["Hello", "World"])),
            ]
        )


def test_invalid_key_assignment():
    storage = Dataset.allocate(size=3)
    with pytest.raises(AssertionError):
        storage["gain_ref_blob/path"] = ["Hello", "World!"]


def test_invalid_key_query():
    storage = Dataset.allocate()
    with pytest.raises(KeyError):
        storage.query({"invalid_field": "Hello world"})


def test_non_existent_key_assignment():
    storage = Dataset.allocate(size=3)
    with pytest.raises(AssertionError):
        storage["gain_ref_blob"] = n.zeros(3)


def test_valid_key_assignment():
    storage = Dataset.allocate(size=3, fields=[("gain_ref_blob/path", "O")])
    storage["gain_ref_blob/path"] = "Hello World!"
    assert isinstance(storage["gain_ref_blob/path"], Column)
    assert len(storage["gain_ref_blob/path"]) == 3


def test_valid_multi_dimensional_key_assignment():
    storage = Dataset.allocate(
        size=3,
        fields=[
            ("location/micrograph_shape", "<u4", (2,)),
            ("alignments3D/shift", "500,2f4"),
        ],
    )

    shapearr = n.array([42, 24])
    shiftarr = n.array([(3.14, 2.67)] * 500, dtype=n.float32)
    storage["location/micrograph_shape"] = shapearr
    storage["alignments3D/shift"] = shiftarr

    assert isinstance(storage["location/micrograph_shape"], Column)
    assert isinstance(storage["alignments3D/shift"], Column)
    assert len(storage["location/micrograph_shape"]) == 3
    assert len(storage["alignments3D/shift"]) == 3
    assert storage["location/micrograph_shape"].shape == (3, 2)
    assert storage["alignments3D/shift"].shape == (3, 500, 2)
    assert n.all(storage["location/micrograph_shape"][2] == shapearr)
    assert n.all(storage["alignments3D/shift"][2] == shiftarr)


def test_add_fields():
    storage = Dataset.allocate(size=2000000, fields=[("field1", "u8"), ("field2", "f4"), ("field3", "O")]).add_fields(
        [
            ("mscope_params/accel_kv", "f4"),
            ("mscope_params/cs_mm", "f4"),
            ("mscope_params/total_dose_e_per_A2", "f4"),
            ("mscope_params/phase_plate", "u4"),
            ("mscope_params/neg_stain", "u4"),
            ("mscope_params/exp_group_id", "u4"),
        ]
    )
    assert len(storage.fields()), 10


def test_add_fields_nonebug(t20s_dset):
    t20s_dset.add_fields([("micrograph_blob_non_dw/path", "O")])
    assert all(t20s_dset["gain_ref_blob/path"] == "J1/imported/norm-amibox05-0.mrc")


def test_to_list():
    storage = Dataset.allocate(size=1, fields=[("field1", "u8"), ("field2", "f4"), ("field3", "O")])
    lst = storage.to_list()
    assert len(lst) == 1
    assert len(lst[0]) == 4


def test_to_list_exclude_uid():
    storage = Dataset.allocate(size=1, fields=[("field1", "u8"), ("field2", "f4"), ("field3", "O")])
    storage["field3"][0] = "Hello"
    lst = storage.to_list(exclude_uid=True)
    assert len(lst) == 1
    assert len(lst[0]) == 3
    assert lst == [[0, 0.0, "Hello"]]


def test_save():
    dtype = [("uid", "u8"), ("field1", "u4"), ("field2", "f4"), ("field3", "S6"), ("field4", "f8")]
    expected = n.array([(1, 42, 3.14, "Hello", 1.0), (2, 42, 2.73, "World", 0.0)], dtype=dtype)
    dset = Dataset(expected)
    new_iodata = BytesIO()
    dset.save(new_iodata)
    new_iodata.seek(0)
    actual = n.load(new_iodata)
    assert expected.dtype.descr == actual.dtype.descr
    assert all(all(n.equal(expected[d[0]], actual[d[0]])) for d in dtype if d[0] != "field3")
    assert all(e.decode() == a.decode() for e, a in zip(expected["field3"], actual["field3"]))


def test_load(io_data):
    dtype = [("field1", "u4"), ("field2", "f4"), ("field3", "O"), ("field4", "f8"), ("field5", "i8")]
    expected = Dataset.allocate(size=2, fields=dtype)
    expected["field1"] = 42
    expected["field2"] = n.array([3.14, 2.73], dtype="f8")
    expected["field3"][:] = n.array(["Hello", "World"])
    expected["field4"][1] = 1.0
    expected["field5"][0:] = 43

    result = Dataset.load(io_data)
    assert expected.descr() == result.descr()
    assert all([n.equal(expected[d[0]], result[d[0]]).all() for d in dtype if d[0] != "uid"])


def test_load_fields(io_data):
    dtype = [("field2", "f4"), ("field3", "O"), ("field5", "i8")]
    expected = Dataset.allocate(size=2, fields=dtype)
    expected["field2"] = n.array([3.14, 2.73], dtype="f8")
    expected["field3"][:] = n.array(["Hello", "World"])
    expected["field5"][0:] = 43

    result = Dataset.load(io_data, fields=["field2", "field3", "field5"])
    assert expected.descr() == result.descr()
    assert all([n.equal(expected[d[0]], result[d[0]]).all() for d in dtype if d[0] != "uid"])


def test_subset_range_out_of_bounds():
    data = Dataset.allocate(size=3, fields=[("field1", "u8"), ("field2", "f4"), ("field3", "O")])
    subset = data.slice(2, 100)
    assert len(subset) == 1


def test_from_data_none():
    data = Dataset()  # FIXME: Not necessary, remove
    assert len(data) == 0


def test_load_stream(small_dset, small_dset_path):
    result = Dataset.load(small_dset_path)
    assert result == small_dset


def test_load_stream_prefixes(small_dset, small_dset_path):
    result = Dataset.load(small_dset_path, prefixes=["field"])
    assert result == small_dset.filter_prefixes(["field"], copy=True)


def test_load_stream_fields(small_dset, small_dset_stream):
    result = Dataset.load(small_dset_stream, fields=["field/2"])
    assert result == small_dset.filter_fields(["field/2"], copy=True)


def test_load_empty_stream(empty_dset, empty_dset_stream):
    result = Dataset.load(empty_dset_stream)
    assert result == empty_dset


def test_pickle_unpickle():
    import pickle

    dset = Dataset(
        [
            ("uid", n.array([1, 2, 3])),
            ("dat", n.array(["Hello", "World", "!"])),
        ]
    )
    pickled = pickle.dumps(dset, protocol=pickle.HIGHEST_PROTOCOL)
    del dset  # calls data destructor to clear dset memory

    dset = pickle.loads(pickled)
    assert n.array_equal(dset["uid"], [1, 2, 3])
    assert n.array_equal(dset["dat"], ["Hello", "World", "!"])


def test_column_aggregation(t20s_dset):
    assert type(t20s_dset["uid"]) is Column
    assert type(n.max(t20s_dset["uid"])) is n.uint64
    assert isinstance(n.mean(t20s_dset["uid"]), n.number)
    assert not isinstance(n.mean(t20s_dset["uid"]), n.ndarray)


@pytest.mark.skipif(n.__version__.startswith("1.17."), reason="works with newer numpy versions, use case is limited")
def test_row_array_type(t20s_dset):
    rowarr = n.array(t20s_dset.rows())
    assert isinstance(rowarr[0], Row)


def test_innerjoin_bigger():
    dat2 = "dat2dat2dat2dat2dat2dat2dat2dat2dat2dat2dat2dat2dat2"
    d1 = Dataset([("uid", [1, 2, 3]), ("dat1", ["Hello", "World", "!"])])
    d2 = Dataset([("uid", [0, 1, 2, 3, 4]), (dat2, ["(", "Hello", "World", "!", ")"])])
    assert d1.innerjoin(d2) == Dataset(
        [
            ("uid", [1, 2, 3]),
            ("dat1", ["Hello", "World", "!"]),
            (dat2, ["Hello", "World", "!"]),
        ]
    )


def test_innerjoin_smaller():
    dat2 = "dat2dat2dat2dat2dat2dat2dat2dat2dat2dat2dat2dat2dat2"
    d1 = Dataset([("uid", [1, 2, 3]), ("dat1", ["Hello", "World", "!"])])
    d2 = Dataset([("uid", [3, 1]), (dat2, ["Hello", "World"])])

    assert d1.innerjoin(d2) == Dataset(
        [
            ("uid", [1, 3]),
            ("dat1", ["Hello", "!"]),
            (dat2, ["World", "Hello"]),
        ]
    )


def test_append_many_empty():
    assert len(Dataset.append_many().rows()) == 0


def test_append_empty_fields():
    d1 = Dataset.allocate(0, [("field", "f4")])
    d2 = Dataset.allocate(0, [("field", "f4")])
    d3 = d1.append(d2)
    assert len(d3) == 0
    assert d3.fields() == ["uid", "field"]


def test_union_many_empty():
    assert len(Dataset.union_many().rows()) == 0


def test_union_empty_fields():
    d1 = Dataset.allocate(0, [("field", "f4")])
    d2 = Dataset.allocate(0, [("field", "f4")])
    d3 = d1.union(d2)
    assert len(d3) == 0
    assert d3.fields() == ["uid", "field"]


def test_allocate_many_separate():
    for _ in range(66_000):
        allocated = []
        for _ in range(3):
            allocated.append(Dataset(1))
        assert len(allocated) == 3
        del allocated


def test_allocate_many_together():
    # Checks for logic issues when allocating a lot of datasets
    for _ in range(3):
        allocated = []
        for _ in range(66_000):
            allocated.append(Dataset(1))
        assert len(allocated) == 66_000
        del allocated


def test_load_4k():
    # Check that a 4kiB dataset (same as linux page size) loads correctly
    # (numpy bug https://github.com/numpy/numpy/pull/27723 fixed in 2.2)
    d = Dataset.load("tests/data/4k_dataset.cs")
    assert len(d) == 0



import sys

from Cython.Build import cythonize  # type: ignore
from setuptools import Extension, setup

DEBUG = False  # set to True to enable debugging
libraries = []
define_macros = []
undef_macros = []
extra_compile_args = []
extra_link_args = []

if sys.platform == "win32":
    extra_compile_args += ["/std:c11"]
    define_macros += [("__WIN32__", 1)]
else:
    libraries.append("pthread")

if sys.platform == "win32" and DEBUG:
    define_macros += [("_DEBUG",)]
    extra_compile_args += ["/Zi"]
    extra_link_args += ["/DEBUG"]
elif DEBUG:
    extra_compile_args += ["-g", "-O0", "-Wall", "-Wextra"]

setup(
    name="cryosparc_tools",
    version="5.0.2",
    description="Toolkit for interfacing with CryoSPARC",
    headers=["cryosparc/include/cryosparc-tools/dataset.h"],
    ext_modules=cythonize(
        Extension(
            name="cryosparc.dataset.core",
            sources=[
                "cryosparc/include/lz4/lib/lz4.c",
                "cryosparc/dataset/dataset.c",
                "cryosparc/dataset/core.pyx",
            ],
            include_dirs=["cryosparc/include/"],
            libraries=libraries,
            define_macros=define_macros,
            undef_macros=undef_macros,
            extra_compile_args=extra_compile_args,
            extra_link_args=extra_link_args,
            depends=[
                "cryosparc/include/lz4/lib/lz4.h",
                "cryosparc/include/cryosparc-tools/dataset.h",
                "cryosparc/dataset/lz4.pxd",
                "cryosparc/dataset/dataset.pxd",
            ],
        ),
        language_level=3,
        gdb_debug=DEBUG,
    ),
)



"""
Functions and classes for performing and storing user authentication operations.
"""

from datetime import datetime, timezone
from functools import lru_cache
from pathlib import Path
from typing import Dict, Optional
from warnings import warn

from pydantic import AwareDatetime, BaseModel, RootModel, ValidationError, field_validator

from .models.auth import Token
from .platform import user_config_path
from .util import first


@lru_cache(maxsize=1)
def get_default_auth_config_path():
    return user_config_path() / "cryosparc-tools" / "auth.json"


class AuthSession(BaseModel):
    """
    A cryosparc-tools CLI session from a successful call to
    ``python -m cryosparc.tools login``.
    """

    token: Token
    expires: AwareDatetime


class UserAuthSessions(RootModel):
    """
    Dictionary of CLI login sessions for a single instance organized by user ID.
    Only one session is allowed per-user.
    """

    root: Dict[str, AuthSession]

    @field_validator("root", mode="after")
    @classmethod
    def remove_expired_tokens(cls, value: Dict[str, AuthSession]):
        now = datetime.now(timezone.utc)
        return {k: v for k, v in value.items() if v.expires > now}


class InstanceAuthSessions(RootModel):
    """
    Dictionary of CLI multiple user sessions for multiple instances,
    organized by instance URL. Stored in ~/.config directory.

    Example:

        >>> {
        ...     "https://cryosparc.example.com": {
        ...         "ali@example.com": {
        ...             "token": {
        ...                 "access_token": "abc123",
        ...                 "token_type": "bearer"
        ...             },
        ...             "expires": "2025-01-31T12:00:00.000000Z"
        ...         }
        ...     }
        ... }
    """

    root: Dict[str, UserAuthSessions]

    @classmethod
    def load(cls, path: Optional[Path] = None):
        """
        Read all auth tokens from the given path. Loads from the
        user config directory if path is unspecified.
        """
        if not path:
            path = get_default_auth_config_path()
        try:
            return cls.model_validate_json(path.read_bytes())
        except FileNotFoundError:
            if path != get_default_auth_config_path():
                warn(f"Sessions file at {path} does not exist; load result is empty")
            return cls({})
        except ValidationError as err:
            warn(str(err))
            warn(f"Sessions stored at {path} are invalid; load result is empty")
            return cls({})

    def save(self, path: Optional[Path] = None):
        """
        Write all auth tokens to the given path. Writes to the user config
        directory if path is unspecified.
        """
        if not path:
            path = get_default_auth_config_path()
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(self.model_dump_json(indent=4))

    def find(self, url: str, email: Optional[str] = None) -> Optional[AuthSession]:
        """
        Find the first available token for the given instance URL and email.
        If an email is not specified, returns the first existing session
        """
        if url in self.root:
            user_sessions = self.root[url]
            return first(sess for e, sess in user_sessions.root.items() if not email or e == email)

    def insert(self, url: str, email: str, token: Token, expires: AwareDatetime):
        """
        Add or replace a new token for the given instance and user email.
        """
        if url not in self.root:
            self.root[url] = UserAuthSessions({})
        self.root[url].root[email] = AuthSession(token=token, expires=expires)



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
import datetime
from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel, Field


class GenerateIntermediateResultsSettings(BaseModel):
    """
    Settings for generating intermediate results for different job types in a project.
    """

    class_2D_new: bool = False
    """
    """
    class_3D: bool = False
    """
    """
    var_3D_disp: bool = False
    """
    """


class ProjectLastAccessed(BaseModel):
    """
    Record for when a user last accessed a project.
    """

    name: str = ""
    """
    Username that last accessed the project.
    """
    accessed_at: datetime.datetime = datetime.datetime(1, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    Timestamp of the last access.
    """


class ProjectStats(BaseModel):
    """
    Statistics about the contents of a project.
    """

    workspace_count: int = 0
    """
    """
    session_count: int = 0
    """
    """
    job_count: int = 0
    """
    """
    job_types: Dict[str, int] = {}
    """
    Number of jobs per type.
    """
    job_sections: Dict[str, int] = {}
    """
    Number of jobs per section.
    """
    job_status: Dict[str, int] = {}
    """
    Number of jobs per status.
    """
    updated_at: Optional[datetime.datetime] = None
    """
    """


class ProjectWorkflowInfo(BaseModel):
    """
    Information about workflows used in a project.
    """

    latest_workflow_uid: str = "WF1"
    """
    UID of the latest workflow applied in a project.
    """
    runs: Dict[str, int] = {}
    """
    Map from workflow ID to the number of runs of that workflow.
    """


class Project(BaseModel):
    """
    Representation of a CryoSPARC Project, a container of data-processing
    results.
    """

    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    dumped_at: Optional[datetime.datetime] = None
    """
    When the model was last dumped to disk
    """
    last_dumped_version: Optional[str] = None
    """
    The version of CryoSPARC last dumped at
    """
    autodump: bool = True
    """
    Whether the model was updated recently and must be dumped

    :meta private:
    """
    uid: str
    """
    Unique identifier for the project, e.g. "P1".
    """
    project_dir: str
    """
    Absolute path to the project directory on disk.
    """
    owner_user_id: str
    """
    User ID of the project owner.
    """
    title: str
    """
    User-specified project title.
    """
    description: str = ""
    """
    User-specified detailed project description.
    """
    project_params_pdef: Dict[str, Any] = {}
    """
    Default job parameter definitions to be used within the project.
    For example ``{'output_f16': True}`` to enable half-precision output for
    all motion correction jobs.
    """
    queue_paused: bool = False
    """
    Currently unused :meta private:
    """
    deleted: bool = False
    """
    """
    deleting: bool = False
    """
    """
    users_with_access: List[str] = []
    """
    List of user IDs with access to the project, including the owner.
    """
    size: int = 0
    """
    Size of the project on disk, in bytes.
    """
    size_last_updated: datetime.datetime
    """
    """
    last_accessed: Optional[ProjectLastAccessed] = None
    """
    """
    archived: bool = False
    """
    """
    detached: bool = False
    """
    """
    generate_intermediate_results_settings: GenerateIntermediateResultsSettings = GenerateIntermediateResultsSettings()
    """
    """
    last_exp_group_id_used: Optional[int] = None
    """
    Last exposure group ID used in the project.
    """
    develop_run_as_user: Optional[str] = None
    """
    :meta private:
    """
    imported_at: Optional[datetime.datetime] = None
    """
    """
    import_status: Optional[Literal["importing", "complete", "failed"]] = None
    """
    """
    project_stats: ProjectStats = ProjectStats()
    """
    """
    created_at_version: str = "unknown"
    """
    """
    last_archived_version: Optional[str] = None
    """
    """
    last_detached_version: Optional[str] = None
    """
    """
    is_cleanup_in_progress: bool = False
    """
    """
    tags: List[str] = []
    """
    """
    starred_by: List[str] = []
    """
    List of user IDs who have starred the project.
    """
    autodump_failed: bool = False
    """
    Whether the last automatic disk dump operation failed.
    """
    autodump_errors: List[str] = []
    """
    List of error messages from the last automatic disk dump operation.
    """
    workflows: Optional[ProjectWorkflowInfo] = None
    """
    """
    uid_num: int
    """
    Numeric part of the project UID.
    """


class ProjectSymlink(BaseModel):
    """
    Information about a symlink in a project directory.
    """

    path: str
    """
    Path of the symlink.
    """
    target: str
    """
    Target of the symlink.
    """
    exists: bool
    """
    Whether the target of the symlink exists.
    """



"""
Utilities and type definitions for working with dataset fields and column types.
"""

import json
from typing import TYPE_CHECKING, Dict, List, Literal, Optional, Sequence, Type, TypedDict, Union

import numpy as n

from ..errors import DatasetLoadError
from ..spec import DType, Field
from .core import Data, DsetType

if TYPE_CHECKING:
    from numpy.typing import DTypeLike, NDArray


class DatasetHeader(TypedDict):
    """
    Encoded dataset file description.
    """

    length: int
    """Number of rows in the dataset"""

    dtype: List[Field]
    """
    Column description
    """

    compression: Literal["lz4", None]
    """Compression library used for in the dataset"""

    compressed_fields: List[str]
    """Field names that require decompression."""


DSET_TO_TYPE_MAP: Dict[DsetType, Type[Union[n.number, n.object_]]] = {
    DsetType.T_F32: n.float32,
    DsetType.T_F64: n.float64,
    DsetType.T_C32: n.complex64,
    DsetType.T_C64: n.complex128,
    DsetType.T_I8: n.int8,
    DsetType.T_I16: n.int16,
    DsetType.T_I32: n.int32,
    DsetType.T_I64: n.int64,
    DsetType.T_U8: n.uint8,
    DsetType.T_U16: n.uint16,
    DsetType.T_U32: n.uint32,
    DsetType.T_STR: n.uint64,  # Note: Prefer T_OBJ when working in Python
    DsetType.T_U64: n.uint64,
    DsetType.T_OBJ: n.object_,
}

TYPE_TO_DSET_MAP = {
    **{v: k for k, v in DSET_TO_TYPE_MAP.items()},
    **{
        float: DsetType.T_F64,
        complex: DsetType.T_C64,
        int: DsetType.T_I64,
        str: DsetType.T_OBJ,
        object: DsetType.T_OBJ,
    },
}

# Set of dataset fields that should not be compressed when saving in
# NEWEST_FORMAT
NEVER_COMPRESS_FIELDS = {"uid"}


def normalize_field(name: str, dtype: "DTypeLike") -> Field:
    # Note: field name "uid" is always uint64, regardless of given dtype
    # Note: sd
    dt = n.dtype(dtype)
    if name == "uid":
        return name, n.dtype(n.uint64).str
    elif dt.char in {"O", "S", "U"}:  # all python string object types
        return name, n.dtype(object).str
    elif dt.shape:
        return name, dt.base.str, dt.shape
    else:
        return name, dt.str


def fielddtype(field: Field) -> DType:
    _, dt, *rest = field
    return (dt, rest[0]) if rest else dt


def arraydtype(a: "NDArray") -> DType:
    assert len(a.dtype.descr) == 1, "Cannot get dtype from record array"
    return (a.dtype.str, a.shape[1:]) if len(a.shape) > 1 else a.dtype.str


def dtypestr(dtype: "DTypeLike") -> str:
    dt = n.dtype(dtype)
    if dt.shape:
        shape = ",".join(map(str, dt.shape))
        return f"{dt.base.str[0]}{shape}{dt.base.str[1:]}"
    else:
        return dt.str


def get_data_field(data: Data, field: str) -> Field:
    return normalize_field(field, get_data_field_dtype(data, field))


def get_data_field_dtype(data: Data, field: str) -> "DTypeLike":
    t = data.type(field)
    if t == 0:
        raise KeyError(f"Unknown dataset field {field}")
    elif t not in DSET_TO_TYPE_MAP:
        raise KeyError(f"Unknown dataset field type {t}")
    dt = n.dtype(DSET_TO_TYPE_MAP[t])
    shape = data.getshp(field)
    return (dt.str, shape) if shape else dt.str


def filter_descr(
    descr: List[Field],
    *,
    keep_prefixes: Optional[Sequence[str]] = None,
    keep_fields: Optional[Sequence[str]] = None,
) -> List[Field]:
    # Get a filtered list of fields based on the user-specified prefixies
    # and/or fields. Returns all fields if no filter params are specified.
    # Always returns at least uid field, if it exists.
    filtered: List[Field] = []
    for field in descr:
        if (
            field[0] == "uid"
            or (keep_prefixes is None and keep_fields is None)
            or (keep_prefixes is not None and any(field[0].startswith(f"{p}/") for p in keep_prefixes))
            or (keep_fields is not None and field[0] in keep_fields)
        ):
            filtered.append(field)
    return filtered


def encode_dataset_header(fields: DatasetHeader) -> bytes:
    return json.dumps(fields).encode()


def decode_dataset_header(data: Union[bytes, dict]) -> DatasetHeader:
    try:
        header = json.loads(data) if isinstance(data, bytes) else data
        assert isinstance(header, dict), f"Incorrect decoded header type (expected dict, got {type(header)})"
        assert "length" in header and isinstance(header["length"], int), (
            'Dataset header "length" key missing or has incorrect type'
        )
        assert "dtype" in header and isinstance(header["dtype"], list), (
            'Dataset header "dtype" key missing or has incorrect type'
        )
        assert "compression" in header and header["compression"] in {
            None,
            "lz4",
        }, 'Dataset header "compression" key missing or has incorrect type'
        assert "compressed_fields" in header or isinstance(header["compressed_fields"], list), (
            'Dataset header "compressed_fields" key missing or has incorrect type'
        )

        length: int = header["length"]
        dtype: List[Field] = [(f, d, tuple(rest[0])) if rest else (f, d) for f, d, *rest in header["dtype"]]
        compression: Literal["lz4", None] = header["compression"]
        compressed_fields: List[str] = header["compressed_fields"]

        return DatasetHeader(
            length=length,
            dtype=dtype,
            compression=compression,
            compressed_fields=compressed_fields,
        )
    except Exception as e:
        raise DatasetLoadError(
            f"Incorrect dataset field format: {data.decode() if isinstance(data, bytes) else data}"
        ) from e



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
import datetime
from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel, Field

from .session import ExposureGroup, LiveComputeResources
from .session_params import LivePreprocessingParams


class RunConfiguration(BaseModel):
    """
    Runtime configuration for a session.
    """

    exposure_processing_priority: Literal["normal", "oldest", "latest", "alternate"] = "normal"
    """
    """
    phase_one_wait_for_exposures: bool = False
    """
    Whether to wait for exposures to become available before launching the Live workers.
    """
    auto_pause: Literal["disabled", "graceful", "immediate"] = "disabled"
    """
    Auto-pause configuration for the session.
    """
    auto_pause_after_idle_minutes: int = 10
    """
    Number of idle minutes before auto-pausing the session. Only used if
    ``auto_pause`` is set to ``"enabled"``.
    """


class SessionConfigProfile(BaseModel):
    """
    A profile for session configuration that can be reused across multiple sessions.
    """

    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    title: str
    """
    """
    created_by_user_id: str
    """
    """
    last_applied_at: Optional[datetime.datetime] = None
    """
    """
    compute_resources: Optional[LiveComputeResources] = None
    """
    """
    exp_groups: List[ExposureGroup] = []
    """
    """
    session_params: Dict[str, Any] = {}
    """
    """
    run_configuration: Optional[RunConfiguration] = None
    """
    """


class SessionConfigProfileBody(BaseModel):
    """
    Required fields to create or update a session configuration profile.
    """

    title: str
    """
    """
    compute_resources: Optional[LiveComputeResources] = None
    """
    """
    exp_groups: List[ExposureGroup] = []
    """
    """
    session_params: Optional[LivePreprocessingParams] = None
    """
    """
    run_configuration: Optional[RunConfiguration] = None
    """
    """



import warnings
from pathlib import PurePath, PurePosixPath
from typing import IO, TYPE_CHECKING, Any, Dict, Iterable, List, Literal, Optional, Tuple, Union

from typing_extensions import Unpack

from ..dataset import DEFAULT_FORMAT, Dataset
from ..dataset.row import R
from ..models.project import Project
from ..search import In, JobSearch
from ..spec import Datatype, SlotSpec
from ..util import PurePosixPathProperty
from . import Controller, as_output_slot
from .job import ExternalJobController, FileOrFigure, JobController
from .workspace import WorkspaceController

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from ..tools import CryoSPARC


class ProjectController(Controller[Project]):
    """
    Accessor instance for CryoSPARC projects with ability to add workspaces, jobs
    and upload/download project files. Should be created with
    :py:meth:`cs.find_project() <cryosparc.tools.CryoSPARC.find_project>`.

    Arguments:
        project (str | Project): either Project UID or Project model, e.g. ``"P3"``

    Attributes:
        model (Project): All project data from the CryoSPARC database. Contents
            may change over time, use :py:meth:`refresh` to update.
    """

    uid: str
    """
    Project unique ID, e.g., "P3"
    """

    def __init__(self, cs: "CryoSPARC", project: Union[str, Project]) -> None:
        self.cs = cs
        if isinstance(project, str):
            self.uid = project
            self.refresh()
        else:
            self.uid = project.uid
            self.model = project

    def refresh(self):
        """
        Reload this project from the CryoSPARC database.

        Returns:
            ProjectController: self
        """
        self.model = self.cs.api.projects.find_one(self.uid)
        return self

    @property
    def title(self) -> str:
        """Project title"""
        return self.model.title

    @property
    def desc(self) -> str:
        """Project description"""
        return self.model.description

    @property
    def dir(self) -> PurePosixPath:
        """Full path to project directory."""
        return PurePosixPathProperty(self.cs.api.projects.get_directory(self.uid))

    def set_title(self, title: str):
        """
        Set the project title.

        Args:
            title (str): New project title
        """
        self.model = self.cs.api.projects.set_title(self.uid, title=title)

    def set_description(self, desc: str):
        """
        Set the project description. May include Markdown formatting.

        Args:
            desc (str): New project description
        """
        self.model = self.cs.api.projects.set_description(self.uid, description=desc)

    def find_workspaces(self, *, order: Literal[1, -1] = 1) -> Iterable[WorkspaceController]:
        """
        Get all workspaces available in the current project.

        Returns:
            Iterable[WorkspaceController]: workspace accessor objects
        """
        return self.cs.find_workspaces(self.uid, order=order)

    def find_workspace(self, workspace_uid: str) -> WorkspaceController:
        """
        Get a workspace accessor instance for the workspace in this project
        with the given UID. Fails with an error if workspace does not exist.

        Args:
            workspace_uid (str): Workspace unique ID, e.g., "W1"

        Returns:
            WorkspaceController: workspace accessor object
        """
        return WorkspaceController(self.cs, (self.uid, workspace_uid))

    def find_jobs(
        self,
        workspace_uid: Optional[In[str]] = None,
        *,
        order: Literal[1, -1] = 1,
        **search: Unpack[JobSearch],
    ) -> Iterable[JobController]:
        """
        Search jobs available in the current project.

        Example:
            >>> jobs = project.find_jobs("W3")
            >>> jobs = project.find_jobs(["W3", "W4"])
            >>> jobs = project.find_jobs(
            ...     type="homo_reconstruct",
            ...     completed_at=(datetime(2025, 3, 1), datetime(2025, 3, 31)),
            ...     order=-1,
            ... )
            >>> for job in jobs:
            ...     print(job.uid)

        Args:
            workspace_uid (str | list[str] | None): Workspace unique ID, e.g.,
                "W1". If not specified, returns jobs from all workspaces.
                Defaults to None.
            **search (JobSearch): Additional search parameters to filter jobs,
                specified as keyword arguments.

        Returns:
            Iterable[JobController]: job accessor objects
        """
        return self.cs.find_jobs(self.uid, workspace_uid, order=order, **search)

    def find_job(self, job_uid: str) -> JobController:
        """
        Get a job accessor instance for the job in this project with the given
        UID. Fails with an error if job does not exist.

        Args:
            job_uid (str): Job unique ID, e.g., "J42"

        Returns:
            JobController: job accessor instance
        """
        return JobController(self.cs, (self.uid, job_uid))

    def find_external_job(self, job_uid: str) -> ExternalJobController:
        """
        Get the External job accessor instance for an External job in this
        project with the given UID. Fails if the job does not exist or is not an
        external job.

        Args:
            job_uid (str): Job unique ID, e.g,. "J42"

        Raises:
            TypeError: If job is not an external job

        Returns:
            ExternalJobController: external job accessor object
        """
        return self.cs.find_external_job(self.uid, job_uid)

    def create_workspace(self, title: str, desc: Optional[str] = None) -> WorkspaceController:
        """
        Create a new empty workspace in this project. At least a title must be
        provided.

        Args:
            title (str): Title of new workspace
            desc (str, optional): Markdown text description. Defaults to None.

        Returns:
            WorkspaceController: created workspace accessor object

        Raises:
            APIError: Workspace cannot be created.
        """
        return self.cs.create_workspace(self.uid, title, desc)

    def create_job(
        self,
        workspace_uid: str,
        type: str,
        connections: Dict[str, Union[Tuple[str, str], List[Tuple[str, str]]]] = {},
        params: Dict[str, Any] = {},
        title: str = "",
        desc: str = "",
    ) -> JobController:
        """
        Create a new job with the given type. Use
        :py:attr:`cs.job_register <cryosparc.tools.CryoSPARC.job_register>`
        to find available job types on the connected CryoSPARC instance.

        Args:
            project_uid (str): Project UID to create job in, e.g., "P3"
            workspace_uid (str): Workspace UID to create job in, e.g., "W1"
            type (str): Job type identifier, e.g., "homo_abinit"
            connections (dict[str, tuple[str, str] | list[tuple[str, str]]]):
                Initial input connections. Each key is an input name and each
                value is a (job uid, output name) tuple. Defaults to {}
            params (dict[str, any], optional): Specify parameter values.
                Defaults to {}.
            title (str, optional): Job title. Defaults to "".
            desc (str, optional): Job markdown description. Defaults to "".

        Returns:
            JobController: created job accessor object.

        Raises:
            APIError: Job cannot be created.

        Examples:

            Create an Import Movies job.

            >>> from cryosparc.tools import CryoSPARC
            >>> cs = CryoSPARC("http://localhost:61000")
            >>> project = cs.find_project("P3")
            >>> import_job = project.create_job("W1", "import_movies")
            >>> import_job.set_param("blob_paths", "/bulk/data/t20s/*.tif")
            True

            Create a 3-class ab-initio job connected to existing particles.

            >>> abinit_job = project.create_job("W1", "homo_abinit"
            ...     connections={"particles": ("J20", "particles_selected")}
            ...     params={"abinit_K": 3}
            ... )
        """
        return self.cs.create_job(
            self.uid, workspace_uid, type, connections=connections, params=params, title=title, desc=desc
        )

    def create_external_job(
        self,
        workspace_uid: str,
        title: str = "",
        desc: str = "",
    ) -> ExternalJobController:
        """
        Add a new External job to this project to save generated outputs to.

        Args:
            workspace_uid (str): Workspace UID to create job in, e.g., "W3".
            title (str, optional): Title for external job (recommended).
                Defaults to "".
            desc (str, optional): Markdown description for external job.
                Defaults to "".

        Returns:
            ExternalJob: created external job instance
        """
        return self.cs.create_external_job(self.uid, workspace_uid=workspace_uid, title=title, desc=desc)

    def save_external_result(
        self,
        workspace_uid: Optional[str],
        dataset: Dataset[R],
        type: Datatype,
        name: Optional[str] = None,
        slots: Optional[List[SlotSpec]] = None,
        passthrough: Optional[Tuple[str, str]] = None,
        title: str = "",
        desc: str = "",
        image: Optional[FileOrFigure] = None,
        savefig_kw: dict = dict(bbox_inches="tight", pad_inches=0),
    ) -> str:
        """
        Save the given result dataset to the project. Specify at least the
        dataset to save and the type of data.

        If ``workspace_uid`` or ``passthrough`` input is not specified or is
        None, saves the result to the project's newest workspace. If
        ``passthrough`` is specified but ``workspace_uid`` is not, saves to the
        passthrough job's newest workspace.

        Returns UID of the External job where the results were saved.

        Examples:

            Save all particle data

            >>> particles = Dataset()
            >>> project.save_external_result("W1", particles, 'particle')
            "J43"

            Save new particle locations that inherit passthrough slots from a
            parent job

            >>> particles = Dataset()
            >>> project.save_external_result(
            ...     workspace_uid='W1',
            ...     dataset=particles,
            ...     type='particle',
            ...     name='particles',
            ...     slots=['location'],
            ...     passthrough=('J42', 'selected_particles'),
            ...     title='Re-centered particles'
            ... )
            "J44"

            Save a result with multiple slots of the same type.

            >>> project.save_external_result(
            ...     workspace_uid="W1",
            ...     dataset=particles,
            ...     type="particle",
            ...     name="particle_alignments",
            ...     slots=[
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_0", "required": True},
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_1", "required": True},
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_2", "required": True},
            ...     ]
            ... )
            "J45"

        Args:
            workspace_uid (str | None): Workspace UID to save results into.
                Specify ``None`` to auto-select a workspace.
            dataset (Dataset): Result dataset.
            type (Datatype): Type of output dataset.
            name (str, optional): Name of output on created External job. Same
                as type if unspecified. Defaults to None.
            slots (list[SlotSpec], optional): List of slots expected to
                be created for this output such as ``location`` or ``blob``. Do
                not specify any slots that were passed through from an input
                unless those slots are modified in the output. Defaults to None.
            passthrough (tuple[str, str], optional): Indicates that this output
                inherits slots from the specified output. e.g., ``("J1",
                "particles")``. Defaults to None.
            title (str, optional): Human-readable title for this output.
                Defaults to "".
            desc (str, optional): Markdown description for this output. Defaults
                to "".
            image (str | Path | IO | Figure, optional): Optional image file
                or matplotlib Figure to set as the image for this output.
                Defaults to None.
            savefig_kw (dict, optional): Additional keyword arguments to pass
                to ``figure.savefig()`` when saving matplotlib Figures. Defaults
                to ``dict(bbox_inches="tight", pad_inches=0)``.

        Returns:
            str: UID of created job where this output was saved
        """
        if slots and any(isinstance(s, dict) and "prefix" in s for s in slots):
            warnings.warn("'prefix' slot key is deprecated. Use 'name' instead.", DeprecationWarning, stacklevel=2)
            slots = [as_output_slot(slot) for slot in slots]  # type: ignore
        return self.cs.save_external_result(
            self.uid,
            workspace_uid,
            dataset=dataset,
            type=type,
            name=name,
            slots=slots,
            passthrough=passthrough,
            title=title,
            desc=desc,
            image=image,
            savefig_kw=savefig_kw,
        )

    def list_files(self, prefix: Union[str, PurePosixPath] = "", recursive: bool = False) -> List[str]:
        """
        Get a list of files inside the project directory.

        Args:
            prefix (str | Path, optional): Subdirectory inside project to list.
                Defaults to "".
            recursive (bool, optional): If True, lists files recursively.
                Defaults to False.

        Returns:
            list[str]: List of file paths relative to the project directory.
        """
        return self.cs.list_files(self.uid, prefix=prefix, recursive=recursive)

    def download(self, path: Union[str, PurePosixPath]):
        """
        Open a file in the current project for reading. Use to get files from a
        remote CryoSPARC instance where the project directory is not available
        on the client file system.

        Args:
            path (str | Path): Name or path of file in project directory.

        Yields:
            HTTPResponse: Use a context manager to read the file from the
            request body.

        Examples:

            Download a project's metadata

            >>> cs = CryoSPARC("http://localhost:61000")
            >>> project = cs.find_project("P3")
            >>> with project.download("project.json") as res:
            >>>     project_data = json.loads(res.read())

        """
        return self.cs.download(self.uid, path)

    def download_file(self, path: Union[str, PurePosixPath], target: Union[str, PurePath, IO[bytes]] = ""):
        """
        Download a file from the project directory to the given target path or
        writeable file handle.

        Args:
            path (str | Path): Name or path of file in project directory.
            target (str | Path | IO, optional): Local file path, directory path or
                writeable file handle to write response data. If not specified,
                downloads to current working directory with same file name.
                Defaults to "".

        Returns:
            Path | IO: resulting target path or file handle.
        """
        return self.cs.download_file(self.uid, path, target)

    def download_dataset(self, path: Union[str, PurePosixPath]):
        """
        Download a .cs dataset file from the given relative path in the project
        directory.

        Args:
            path (str | Path): Name or path to .cs file in project directory.

        Returns:
            Dataset: Loaded dataset instance
        """
        return self.cs.download_dataset(self.uid, path)

    def download_mrc(self, path: Union[str, PurePosixPath]):
        """
        Download a .mrc file from the given relative path in the project
        directory.

        Args:
            path (str | Path): Name or path to .mrc file in project directory.

        Returns:
            tuple[Header, NDArray]: MRC file header and data as a numpy array
        """
        return self.cs.download_mrc(self.uid, path)

    def upload(
        self,
        target_path: Union[str, PurePosixPath],
        source: Union[str, bytes, PurePath, IO],
        *,
        overwrite: bool = False,
    ):
        """
        Upload the given file to the project directory at the given relative
        path. Fails if target already exists.

        Args:
            target_path (str | Path): Name or path of file to write in project
                directory.
            source (str | bytes | Path | IO): Local path or file handle to
                upload. May also specified as raw bytes.
            overwrite (bool, optional): If True, overwrite existing files.
                Defaults to False.
        """
        return self.cs.upload(self.uid, target_path, source, overwrite=overwrite)

    def upload_dataset(
        self,
        target_path: Union[str, PurePosixPath],
        dset: Dataset,
        *,
        format: int = DEFAULT_FORMAT,
        overwrite: bool = False,
    ):
        """
        Upload a dataset as a CS file into the project directory. Fails if
        target already exists.

        Args:
            target_path (str | Path): Name or path of dataset to save in the
                project directory. Should have a ``.cs`` extension.
            dset (Dataset): Dataset to save.
            format (int): Format to save in from ``cryosparc.dataset.*_FORMAT``,
                defaults to NUMPY_FORMAT)
            overwrite (bool, optional): If True, overwrite existing files.
                Defaults to False.
        """
        return self.cs.upload_dataset(self.uid, target_path, dset, format=format, overwrite=overwrite)

    def upload_mrc(
        self,
        target_path: Union[str, PurePosixPath],
        data: "NDArray",
        psize: float,
        *,
        overwrite: bool = False,
    ):
        """
        Upload a numpy 2D or 3D array to the project directory as an MRC file. Fails
        if target already exists.

        Args:
            target_path (str | Path): Name or path of MRC file to save in the
                project directory. Should have a ``.mrc`` extension.
            data (NDArray): Numpy array with MRC file data.
            psize (float): Pixel size to include in MRC header.
            overwrite (bool, optional): If True, overwrite existing files.
                Defaults to False.
        """
        return self.cs.upload_mrc(self.uid, target_path, data, psize, overwrite=overwrite)

    def mkdir(
        self,
        target_path: Union[str, PurePosixPath],
        parents: bool = False,
        exist_ok: bool = False,
    ):
        """
        Create a directory in the given project.

        Args:
            target_path (str | Path): Name or path of folder to create inside
                the project directory.
            parents (bool, optional): If True, any missing parents are created
                as needed. Defaults to False.
            exist_ok (bool, optional): If True, does not raise an error for
                existing directories. Still raises if the target path is not a
                directory. Defaults to False.
        """
        self.cs.mkdir(
            project_uid=self.uid,
            target_path=target_path,
            parents=parents,
            exist_ok=exist_ok,
        )

    def cp(self, source_path: Union[str, PurePosixPath], target_path: Union[str, PurePosixPath] = ""):
        """
        Copy a file or folder into the job direcotry.

        Args:
            source_path (str | Path): Relative or absolute path of source file
                or folder to copy. If relative, assumed to be within the project
                directory.
            target_path (str | Path, optional): Name or path in the project
                directory to copy into. If not specified, uses the same file
                name as the source. Defaults to "".
        """
        self.cs.cp(
            project_uid=self.uid,
            source_path=source_path,
            target_path=target_path,
        )

    def symlink(self, source_path: Union[str, PurePosixPath], target_path: Union[str, PurePosixPath] = ""):
        """
        Create a symbolic link in the given project. May only create links for
        files within the project.

        Args:
            source_path (str | Path): Relative or absolute path of source file
                or folder to create a link to. If relative, assumed to be within
                the project directory.
            target_path (str | Path): Name or path of new symlink in the project
                directory. If not specified, creates link with the same file
                name as the source. Defaults to "".
        """
        self.cs.symlink(
            project_uid=self.uid,
            source_path=source_path,
            target_path=target_path,
        )



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
from typing import Literal, Optional

from pydantic import BaseModel


class SchedulerLane(BaseModel):
    """
    Lane definition for the Job scheduler. Scheduler targets are associated with
    a lane that has the same type.
    """

    name: str
    """
    Identifier for this lane.
    """
    type: Literal["node", "cluster"]
    """
    What kind of lane this is based on how on what kind of target(s) it contains.
    """
    title: str
    """
    Human-readable lane title.
    """
    desc: Optional[str] = None
    """
    Optional lane description
    """



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
from typing import Optional

from pydantic import BaseModel

from .job_spec import OutputRef, OutputSpec


class ExternalOutputSpec(BaseModel):
    """
    Specification for an external job with a single output.
    """

    name: str
    """
    Name of output to create on external job.
    """
    spec: OutputSpec
    """
    Specification of output.
    """
    connection: Optional[OutputRef] = None
    """
    Optional source job UID and output to connect and passthrough unmodified
    results to the output. Must have the same type as in the output spec.
    Input name to create and connect should be defined in ``spec.passthrough``,
    otherwise will default to ``connection.output``.
    """



import os
import sys
from functools import lru_cache
from pathlib import Path
from typing import Callable

__all__ = ("is_localhost", "user_config_path")


def is_localhost(hostname: str) -> bool:
    """
    Check if a hostname refers to the current host.
    """
    import socket

    if hostname == "localhost" or hostname.startswith("127.") or hostname == "::1":
        return True
    if hostname == socket.gethostname().replace(" ", "") or hostname == socket.getfqdn().replace(" ", ""):
        return True
    try:
        ip = socket.gethostbyname(hostname).strip()
        return ip.startswith("127.") or ip == "::1"
    except Exception:
        return False


def user_config_path() -> Path:
    """
    config directory tied to the user, e.g. ``~/.config`` or ``$XDG_CONFIG_HOME``
    """
    if sys.platform == "win32":
        path = os.path.normpath(get_win_appdata_folder())
    else:
        path = os.environ.get("XDG_CONFIG_HOME", "")
        if not path.strip():
            path = os.path.expanduser("~/.config")
    return Path(path)


# Windows platform directory code code adapted from platformdirs package
# https://github.com/tox-dev/platformdirs/blob/4.3.6/src/platformdirs/windows.py
# MIT License
# Copyright (c) 2010-202x The platformdirs developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


def get_win_appdata_folder_from_env_vars() -> str:
    """Get folder from environment variables."""
    env_var_name = "APPDATA"
    result = os.environ.get(env_var_name)
    if result is None:
        raise ValueError(f"Unset environment variable: {env_var_name}")
    return result


def get_win_appdata_folder_from_registry() -> str:
    """
    Get folder from the registry.

    This is a fallback technique at best. I'm not sure if using the registry for these guarantees us the correct answer
    for all CSIDL_* names.

    """
    shell_folder_name = "AppData"
    if sys.platform != "win32":
        # only needed for type checker to know that this code runs only on Windows
        raise NotImplementedError

    import winreg

    key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, r"Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders")
    directory, _ = winreg.QueryValueEx(key, shell_folder_name)
    return str(directory)


def get_win_appdata_folder_via_ctypes() -> str:
    """Get folder with ctypes."""
    import ctypes

    csidl_const = 26  # for APPDATA
    buf = ctypes.create_unicode_buffer(1024)
    windll = getattr(ctypes, "windll")
    windll.shell32.SHGetFolderPathW(None, csidl_const, None, 0, buf)

    # Downgrade to short path name if it has high-bit chars.
    if any(ord(c) > 255 for c in buf):
        buf2 = ctypes.create_unicode_buffer(1024)
        if windll.kernel32.GetShortPathNameW(buf.value, buf2, 1024):
            buf = buf2
    return buf.value


def _pick_get_win_appdata_folder() -> Callable[[], str]:
    try:
        import ctypes
    except ImportError:
        pass
    else:
        if hasattr(ctypes, "windll"):
            return get_win_appdata_folder_via_ctypes
    try:
        import winreg  # noqa: F401
    except ImportError:
        return get_win_appdata_folder_from_env_vars
    else:
        return get_win_appdata_folder_from_registry


get_win_appdata_folder = lru_cache(maxsize=None)(_pick_get_win_appdata_folder())



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
from typing import List, Literal, Optional

from pydantic import BaseModel


class DeleteJobPreview(BaseModel):
    """
    Job from a delete request that will be deleted or unlinked.
    """

    project_uid: str
    """
    """
    uid: str
    """
    """
    workspace_uids: List[str]
    """
    """
    status: str
    """
    """
    title: str
    """
    """
    type: str
    """
    """


class DeleteProjectWorkspacePreview(BaseModel):
    """
    Preview of a workspace within a project to be deleted.
    """

    project_uid: str
    """
    """
    uid: str
    """
    """
    title: Optional[str]
    """
    """


class DeleteProjectPreview(BaseModel):
    """
    Preview of a project delete operation, including jobs and workspaces that will be deleted.
    """

    jobs: List[DeleteJobPreview]
    """
    """
    workspaces: List[DeleteProjectWorkspacePreview]
    """
    """


class KeepJobPreview(BaseModel):
    """
    Job from a delete request that cannot be deleted and the reasons for it.
    """

    project_uid: str
    """
    """
    uid: str
    """
    """
    workspace_uids: List[str]
    """
    """
    status: str
    """
    """
    title: str
    """
    """
    type: str
    """
    """
    reason: Literal["final", "descendants"]
    """
    Reason for not deleting this job, either "final" because it or one if its
    descendants is marked a final, or "descendants" because it has children that
    are not or cannot be deleted.
    """
    descendant_job_uids: List[str] = []
    """
    This job's descendant UIDs that cannot be deleted.
    """
    descendant_workspace_uids: List[str] = []
    """
    Workspaces that contain this job's descendants that cannot be deleted.
    """


class DeleteWorkspacePreview(BaseModel):
    """
    Preview of a workspace delete operation, including jobs that will be deleted
    or unlinked and those that will be kept. If there any kept jobs, the
    workspace will not be marked as deleted.
    """

    delete: List[DeleteJobPreview]
    """
    Jobs to be deleted.
    """
    unlink: List[DeleteJobPreview]
    """
    Jobs to be unlinked.
    """
    keep: List[KeepJobPreview]
    """
    Jobs that cannot be deleted, either because they are final or have descendants.
    """
    jobs: List[DeleteJobPreview]
    """
    """



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
from pydantic import BaseModel


class AppSession(BaseModel):
    """
    Data required to verify a web application session.
    """

    user_id: str
    """
    """
    session_id: str
    """
    """
    signature: str
    """
    """


class SHA256Password(BaseModel):
    """
    SHA-256 hashed password data, in hexadecimal format.
    """

    password: str
    """
    """
